# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2018 Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Provides the base for obtaining and storing covers, also known
as album art.
"""

from gi.repository import GLib
from gi.repository import Gio
import logging
import hashlib
import os
import pickle
from typing import Optional

from xl.nls import gettext as _
from xl import common, event, providers, settings, trax, xdg

logger = logging.getLogger(__name__)


# TODO: maybe this could go into common.py instead? could be
# useful in other areas.
class Cacher:
    """
    Simple on-disk cache.

    Note that as entries are stored as
    individual files, the data being stored should be of significant
    size (several KB) or a lot of disk space will likely be wasted.
    """

    def __init__(self, cache_dir):
        """
        :param cache_dir: directory to use for the cache. will be
            created if it does not exist.
        """
        try:
            os.makedirs(cache_dir)
        except OSError:
            pass
        self.cache_dir = cache_dir

    def add(self, data):
        """
        Adds an entry to the cache.  Returns a key that can be used
        to retrieve the data from the cache.

        :param data: The data to store, as a bytestring.
        """
        # FIXME: this doesnt handle hash collisions at all. with
        # 2^256 possible keys its unlikely that we'll have a collision,
        # but we should handle it anyway.
        h = hashlib.sha256()
        h.update(data)
        key = h.hexdigest()
        path = os.path.join(self.cache_dir, key)
        with open(path, "wb") as fp:
            fp.write(data)
        return key

    def remove(self, key):
        """
        Remove an entry from the cache.

        :param key: The key to remove data for.
        """
        path = os.path.join(self.cache_dir, key)
        try:
            os.remove(path)
        except OSError:
            pass

    def get(self, key):
        """
        Retrieve an entry from the cache.  Returns None if the given
        key does not exist.

        :param key: The key to retrieve data for.
        """
        path = os.path.join(self.cache_dir, key)
        if os.path.exists(path):
            with open(path, "rb") as fp:
                return fp.read()
        return None


class CoverManager(providers.ProviderHandler):
    """
    Handles finding covers from various sources.
    """

    DB_VERSION = 2

    def __init__(self, location):
        """
        :param location: The directory to load and store data in.
        """
        providers.ProviderHandler.__init__(self, "covers")
        self.__cache = Cacher(os.path.join(location, 'cache'))
        self.location = location
        self.methods = {}
        self.order = settings.get_option('covers/preferred_order', [])
        self.db = {'version': self.DB_VERSION}
        self.load()
        for method in self.get_providers():
            self.on_provider_added(method)

        with open(xdg.get_data_path('images', 'nocover.png'), 'rb') as f:
            self.default_cover_data = f.read()

        self.tag_fetcher = TagCoverFetcher()
        self.localfile_fetcher = LocalFileCoverFetcher()

        if settings.get_option('covers/use_tags', True):
            providers.register('covers', self.tag_fetcher)
        if settings.get_option('covers/use_localfile', True):
            providers.register('covers', self.localfile_fetcher)

        event.add_callback(self._on_option_set, 'covers_option_set')

    def _on_option_set(self, name, obj, data):
        if data == "covers/use_tags":
            if settings.get_option("covers/use_tags"):
                providers.register('covers', self.tag_fetcher)
            else:
                providers.unregister('covers', self.tag_fetcher)
        elif data == "covers/use_localfile":
            if settings.get_option("covers/use_localfile"):
                providers.register('covers', self.localfile_fetcher)
            else:
                providers.unregister('covers', self.localfile_fetcher)

    def _get_methods(self, fixed=False):
        """
        Returns a list of Methods, sorted by preference

        :param fixed: If true, include fixed-position backends in the
                returned list.
        """
        methods = []
        for name in self.order:
            if name in self.methods:
                methods.append(self.methods[name])
        for k, method in self.methods.items():
            if method not in methods:
                methods.append(method)
        nonfixed = [m for m in methods if not m.fixed]
        if fixed:
            fixed = [m for m in methods if m.fixed]
            fixed.sort(key=lambda x: x.fixed_priority)
            for i, v in enumerate(fixed):
                if v.fixed_priority > 50:
                    methods = fixed[:i] + nonfixed + fixed[i:]
                    break
            else:
                methods = fixed + nonfixed
        else:
            methods = nonfixed
        return methods

    @staticmethod
    def _get_track_key(track: trax.Track) -> Optional[str]:
        """Get a unique, hashable identifier for the track's album.

        If the track has no album identifier, this method returns None.
        """

        # The output is in the form
        #   'tag1  \0  value1a \1 value1b  \0  tag2  \0  value2'
        # without the spaces.
        #
        # Possible tag combinations, in order of preference:
        #   * musicbrainz_albumid
        #   * album albumartist [date]
        #   * __compilation [date]
        #   * album [artist] [date]

        def _get_pair(tag: str) -> Optional[str]:
            value = track.get_tag_raw(tag)
            if not value:
                return None
            value = '\1'.join(value)
            return tag + '\0' + value

        albumid = _get_pair('musicbrainz_albumid')
        if albumid:
            return albumid

        album = _get_pair('album')
        if not album:
            return None

        albumartist = _get_pair('albumartist')
        if albumartist:
            dbkey = album + '\0' + albumartist
        else:
            compilation = _get_pair('__compilation')
            if compilation:
                # compilation is directory+album, where the directory mimics
                # the role of albumartist.
                dbkey = compilation
            else:
                dbkey = album
                artist = _get_pair('artist')
                if artist:
                    dbkey += '\0' + artist
        assert dbkey

        date = _get_pair('date')
        if date:
            dbkey += '\0' + date

        return dbkey

    def get_db_string(self, track: trax.Track) -> Optional[str]:
        """
        Returns the internal string used to map the cover
        to a track

        :param track: the track to retrieve the string for
        :type track: :class:`xl.trax.Track`
        :returns: the internal identifier string
        """
        key = self._get_track_key(track)
        if key is None:
            return None

        return self.db.get(key)

    @common.synchronized
    @common.cached(5)
    def find_covers(self, track, limit=-1, local_only=False):
        """
        Find all covers for a track

        :param track: The track to find covers for
        :param limit: maximum number of covers to return. -1=unlimited.
        :param local_only: If True, will only return results from local
                sources.
        """
        if track is None:
            return
        covers = []
        for method in self._get_methods(fixed=True):
            if local_only and method.use_cache:
                continue
            new = method.find_covers(track, limit=limit)
            new = ["%s:%s" % (method.name, x) for x in new]
            covers.extend(new)
            if limit != -1 and len(covers) >= limit:
                break
        return covers

    def set_cover(self, track, db_string, data=None):
        """
        Sets the cover for a track. This will overwrite any existing
        entry.

        :param track: The track to set the cover for
        :param db_string: the string identifying the source of the
                cover, in "method:key" format.
        :param data: The raw cover data to store for the track.  Will
                only be stored if the method has use_cache=True
        """
        name = db_string.split(":", 1)[0]
        method = self.methods.get(name)
        if method and method.use_cache and data:
            db_string = "cache:%s" % self.__cache.add(data)
        key = self._get_track_key(track)
        if key:
            self.db[key] = db_string
            self.timeout_save()
            event.log_event('cover_set', self, track)

    def remove_cover(self, track):
        """
        Remove the saved cover entry for a track, if it exists.
        """
        if track is None:
            return
        key = self._get_track_key(track)
        if key is None:
            return
        db_string = self.get_db_string(track)
        if db_string is None:
            return
        del self.db[key]
        self.__cache.remove(db_string)
        self.timeout_save()
        event.log_event('cover_removed', self, track)

    def get_cover(self, track, save_cover=True, set_only=False, use_default=False):
        """
        get the cover for a given track.
        if the track has no set cover, backends are
        searched until a cover is found or we run out of backends.

        :param track: the Track to get the cover for.
        :param save_cover: if True, a set_cover call will be made
                to store the cover for later use.
        :param set_only: Only retrieve covers that have been set
                in the db.
        :param use_default: If True, returns the default cover instead
                of None when no covers are found.
        """
        if track is None:
            return self.get_default_cover() if use_default else None

        db_string = self.get_db_string(track)
        if db_string:
            cover = self.get_cover_data(db_string, use_default=use_default)
            if cover:
                return cover

        if set_only:
            return self.get_default_cover() if use_default else None

        covers = self.find_covers(track, limit=1)
        if covers:
            cover = covers[0]
            data = self.get_cover_data(cover, use_default=use_default)
            if save_cover and data != self.get_default_cover():
                self.set_cover(track, cover, data)
            return data

        return self.get_default_cover() if use_default else None

    def get_cover_data(self, db_string, use_default=False):
        """
        Get the raw image data for a cover.

        :param db_string: The db_string identifying the cover to get.
        :param use_default: If True, returns the default cover instead
                of None when no covers are found.
        """
        source, data = db_string.split(":", 1)
        ret = None
        if source == "cache":
            ret = self.__cache.get(data)
        else:
            method = self.methods.get(source)
            if method:
                ret = method.get_cover_data(data)
        if ret is None and use_default is True:
            ret = self.get_default_cover()
        return ret

    def get_default_cover(self):
        """
        Get the raw image data for the cover to show if there is no
        cover to display.
        """
        # TODO: wrap this into get_cover_data and get_cover somehow?
        return self.default_cover_data

    def load(self):
        """
        Load the saved db
        """
        path = os.path.join(self.location, 'covers.db')
        data = None
        for loc in [path, path + ".old", path + ".new"]:
            try:
                with open(loc, 'rb') as f:
                    data = pickle.load(f)
            except IOError:
                pass
            except EOFError:
                try:
                    os.remove(loc)
                except Exception:
                    pass
            if data:
                break
        if data:
            self.db = data
        version = self.db.get('version', 1)
        if version > self.DB_VERSION:
            logger.error(
                "covers.db version (%s) higher than supported (%s); using anyway",
                version,
                self.DB_VERSION,
            )

    @common.glib_wait_seconds(60)
    def timeout_save(self):
        self.save()

    def save(self):
        """
        Save the db
        """
        path = os.path.join(self.location, 'covers.db')
        try:
            with open(path + ".new", 'wb') as f:
                pickle.dump(self.db, f, common.PICKLE_PROTOCOL)
        except IOError:
            return
        try:
            os.rename(path, path + ".old")
        except OSError:
            pass  # if it doesn'texist we don't care
        os.rename(path + ".new", path)
        try:
            os.remove(path + ".old")
        except OSError:
            pass

    def on_provider_added(self, provider):
        self.methods[provider.name] = provider
        if provider.name not in self.order:
            self.order.append(provider.name)

    def on_provider_removed(self, provider):
        try:
            del self.methods[provider.name]
        except KeyError:
            pass
        if provider.name in self.order:
            self.order.remove(provider.name)

    def set_preferred_order(self, order):
        """
        Sets the preferred search order

        :param order: a list containing the order you'd like to search
            first
        """
        if not type(order) in (list, tuple):
            raise TypeError("order must be a list or tuple")
        self.order = order
        settings.set_option('covers/preferred_order', list(order))

    def get_cover_for_tracks(self, tracks, db_strings_to_ignore):
        """
        For tracks, try to find a cover
        Basically returns the first cover found
        :param tracks: list of tracks [xl.trax.Track]
        :param db_strings_to_ignore: list [str]
        :return: GdkPixbuf.Pixbuf or None if no cover found
        """
        for track in tracks:
            db_string = self.get_db_string(track)
            if db_string and db_string not in db_strings_to_ignore:
                db_strings_to_ignore.append(db_string)
                return self.get_cover_data(db_string)

        return None  # No cover found


class CoverSearchMethod:
    """
    Base class for creating cover search methods.

    Search methods do not have to inherit from this class, it's
    intended more as a template to demonstrate the needed interface.
    """

    #: If true, cover results will be cached for faster lookup
    use_cache = True
    #: A name uniquely identifing the search method.
    name = "base"
    #: Whether the backend should have a fixed priority instead of being
    #  configurable.
    fixed = False
    #: Priority for fixed-position backends. Lower is earlier, non-fixed
    #  backends will always be 50.
    fixed_priority = 50

    def find_covers(self, track, limit=-1):
        """
        Find the covers for a given track.

        :param track: The track to find covers for.
        :param limit: Maximal number of covers to return.
        :returns: A list of strings that can be passed to get_cover_data.
        """
        raise NotImplementedError

    def get_cover_data(self, db_string):
        """
        Get the image data for a cover

        :param db_string: A method-dependent string that identifies the
                cover to get.
        """
        raise NotImplementedError


class TagCoverFetcher(CoverSearchMethod):
    """
    Cover source that looks for images embedded in tags.
    """

    use_cache = False
    name = "tags"
    title = _('Tags')
    cover_tags = ["cover", "coverart"]
    fixed = True
    fixed_priority = 30

    def find_covers(self, track, limit=-1):
        covers = []
        tagname = None
        uri = track.get_loc_for_io()

        for tag in self.cover_tags:
            try:
                # Force type conversion to list, fails for None
                covers = list(track.get_tag_disk(tag))
                tagname = tag
                break
            except (TypeError, KeyError):
                pass

        return [
            '{tagname}:{index}:{uri}'.format(tagname=tagname, index=index, uri=uri)
            for index in range(0, len(covers))
        ]

    def get_cover_data(self, db_string):
        tag, index, uri = db_string.split(':', 2)
        track = trax.Track(uri, scan=False)
        covers = track.get_tag_disk(tag)

        if not covers:
            return None

        return covers[int(index)].data


class LocalFileCoverFetcher(CoverSearchMethod):
    """
    Cover source that looks for images in the same directory as the
    Track.
    """

    use_cache = False
    name = "localfile"
    title = _('Local file')
    uri_types = ['file', 'smb', 'sftp', 'nfs']
    extensions = ['.png', '.jpg', '.jpeg', '.gif']
    preferred_names = []
    fixed = True
    fixed_priority = 31

    def __init__(self):
        CoverSearchMethod.__init__(self)

        event.add_callback(self.on_option_set, 'covers_localfile_option_set')
        self.on_option_set(
            'covers_localfile_option_set', settings, 'covers/localfile/preferred_names'
        )

    def find_covers(self, track, limit=-1):
        # TODO: perhaps should instead check to see if its mounted in
        # gio, rather than basing this on uri type. file:// should
        # always be checked, obviously.
        if track.get_type() not in self.uri_types:
            return []
        basedir = Gio.File.new_for_uri(track.get_loc_for_io()).get_parent()
        try:
            if (
                not basedir.query_info(
                    "standard::type", Gio.FileQueryInfoFlags.NONE, None
                ).get_file_type()
                == Gio.FileType.DIRECTORY
            ):
                return []
        except GLib.Error:
            return []
        covers = []
        for fileinfo in basedir.enumerate_children(
            "standard::type" ",standard::name", Gio.FileQueryInfoFlags.NONE, None
        ):
            gloc = basedir.get_child(fileinfo.get_name())
            if not fileinfo.get_file_type() == Gio.FileType.REGULAR:
                continue
            filename = gloc.get_basename()
            base, ext = os.path.splitext(filename)
            if ext.lower() not in self.extensions:
                continue
            if base in self.preferred_names:
                covers.insert(0, gloc.get_uri())
            else:
                covers.append(gloc.get_uri())
        if limit == -1:
            return covers
        else:
            return covers[:limit]

    def get_cover_data(self, db_string):
        try:
            data = Gio.File.new_for_uri(db_string).load_contents(None)[1]
            return data
        except GLib.GError:
            return None

    def on_option_set(self, e, settings, option):
        """
        Updates the internal settings upon option change
        """
        if option == 'covers/localfile/preferred_names':
            self.preferred_names = settings.get_option(option, ['album', 'cover'])


#: The singleton :class:`CoverManager` instance
MANAGER = CoverManager(location=xdg.get_data_home_path("covers", check_exists=False))



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging
import os
import random
import time

from xl import xdg, providers, settings
from xl.trax import search

logger = logging.getLogger(__name__)


class DynamicManager(providers.ProviderHandler):
    """
    handles matching of songs for dynamic playlists
    """

    def __init__(self, collection=[]):
        providers.ProviderHandler.__init__(self, "dynamic_playlists")
        self.buffersize = settings.get_option("playback/dynamic_buffer", 5)
        self.collection = collection
        self.cachedir = os.path.join(xdg.get_cache_dir(), 'dynamic')
        if not os.path.exists(self.cachedir):
            os.makedirs(self.cachedir)

    def find_similar_tracks(self, track, limit=-1, exclude=[]):
        """
        finds tracks from the collection that are similar
        to the passed track.

        @param track: the track to find similar tracks to
        @param limit: limit the returned list to this many
            tracks. If there are more tracks than this
            found, a random selection of those tracks is
            returned.
        """
        logger.debug("Searching for %s tracks related to %s", limit, track)
        artists = self.find_similar_artists(track)
        if artists == []:
            return []
        tracks = []
        random.shuffle(artists)
        i = 0
        while (limit > len(tracks) or limit == -1) and i < len(artists):
            artist = artists[i][1].replace('"', '\\\"')
            i += 1
            searchres = search.search_tracks_from_string(
                self.collection, 'artist=="%s"' % artist, case_sensitive=False
            )
            choices = [x.track for x in searchres]
            if choices == []:
                continue
            random.shuffle(choices)
            j = 0
            while j < len(choices):
                track = choices[j]
                if track not in exclude:
                    tracks.append(track)
                    break
                j += 1
        return tracks

    def find_similar_artists(self, track):
        info = self._load_saved_info(track)
        if info == []:
            info = self._query_sources(track)
            self._save_info(track, info)

        return info

    def _query_sources(self, track):
        info = []
        artist = track.get_tag_raw('artist')
        if not artist:
            return info
        for source in self.get_providers():
            sinfo = source.get_results(','.join(artist))
            info += sinfo
        info.sort(reverse=True)  # TODO: merge artists that are the same
        return info

    def _load_saved_info(self, track):
        artist = track.get_tag_raw('artist')
        if not artist:
            return []
        filename = os.path.join(self.cachedir, ','.join(artist))
        if not os.path.exists(filename):
            return []
        with open(filename) as f:
            line = f.readline()
            if line == '':
                return []
            last_update = float(line)
            if 604800 < time.time() - last_update:  # one week
                info = self._query_sources(track)
                if info != []:
                    self._save_info(track, info)
                    return info
            info = []
            for line in f:
                try:
                    rel, artist = line.strip().split(" ", 1)
                    info.append((rel, artist))
                except Exception:
                    pass
        return info

    def _save_info(self, track, info):
        if info == []:
            return
        filename = os.path.join(self.cachedir, track.get_tag_raw('artist', join=True))
        with open(filename, 'w') as f:
            f.write("%s\n" % time.time())
            for item in info:
                f.write("%.2f %s\n" % item)

    def populate_playlist(self, playlist):
        """
        adds tracks to playlists as needed.
        called when the position of a playlist changes.
        """
        if not playlist:
            return
        current_pos = playlist.current_position
        if current_pos < 0 or current_pos >= len(playlist):
            return
        needed = self.buffersize - (len(playlist) - current_pos)

        if needed < 1:
            needed = 1
        curr = playlist.current

        starttime = time.time()
        tracks = self.find_similar_tracks(curr, needed, playlist)

        remainingtime = 5 - (time.time() - starttime)

        if remainingtime > 0:
            time.sleep(remainingtime)

        if playlist.current_position != current_pos:
            return  # we skipped in that 5 seconds, so ignore it
        playlist.extend(tracks)
        logger.debug("Added %s tracks.", len(tracks))


MANAGER = DynamicManager()


class DynamicSource:
    def __init__(self):
        pass

    def get_results(self, artist):
        raise NotImplementedError

    def _set_manager(self, manager):
        self.manager = manager


# vim: et sts=4 sw=4



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from xlgui.preferences import widgets
from xl.nls import gettext as _
import os
from gi.repository import Gtk

name = _('GroupTagger')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'gt_prefs.ui')


def _get_system_default_font():
    return Gtk.Widget.get_default_style().font_desc.to_string()


class GTPanelFontPreference(widgets.FontButtonPreference):
    default = _get_system_default_font()
    name = 'plugin/grouptagger/panel_font'


class GTPanelFontResetButtonPreference(widgets.FontResetButtonPreference):
    default = _get_system_default_font()
    name = 'plugin/grouptagger/reset_button'
    condition_preference_name = 'plugin/grouptagger/panel_font'


class GTGroupingTagName(widgets.ComboPreference):
    default = 'grouping'
    name = 'plugin/grouptagger/tagname'



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
import logging
from gi.repository import Pango
import os.path

from xl import metadata, providers, settings, xdg
from xl.common import clamp
from xl.playlist import (
    is_valid_playlist,
    import_playlist,
    export_playlist,
    InvalidPlaylistTypeError,
    PlaylistExportOptions,
)
from xl.nls import gettext as _

from xlgui.guiutil import GtkTemplate
from threading import Thread

logger = logging.getLogger(__name__)


def error(parent, message=None, markup=None):
    """
    Shows an error dialog
    """
    if message is markup is None:
        raise ValueError("message or markup must be specified")
    dialog = Gtk.MessageDialog(
        buttons=Gtk.ButtonsType.CLOSE,
        message_type=Gtk.MessageType.ERROR,
        modal=True,
        transient_for=parent,
    )
    if markup is None:
        dialog.props.text = message
    else:
        dialog.set_markup(markup)
    dialog.run()
    dialog.destroy()


def info(parent, message=None, markup=None):
    """
    Shows an info dialog
    """
    if message is markup is None:
        raise ValueError("message or markup must be specified")
    dialog = Gtk.MessageDialog(
        buttons=Gtk.ButtonsType.OK,
        message_type=Gtk.MessageType.INFO,
        modal=True,
        transient_for=parent,
    )
    if markup is None:
        dialog.props.text = message
    else:
        dialog.set_markup(markup)
    dialog.run()
    dialog.destroy()


def yesno(parent, message):
    '''Gets a Yes/No response from a user'''
    dlg = Gtk.MessageDialog(
        buttons=Gtk.ButtonsType.YES_NO,
        message_type=Gtk.MessageType.QUESTION,
        text=message,
        transient_for=parent,
    )
    response = dlg.run()
    dlg.destroy()
    return response


@GtkTemplate('ui', 'about_dialog.ui')
class AboutDialog(Gtk.AboutDialog):
    """
    A dialog showing program info and more
    """

    __gtype_name__ = 'AboutDialog'

    def __init__(self, parent=None):
        Gtk.AboutDialog.__init__(self)
        self.init_template()

        self.set_transient_for(parent)
        logo = GdkPixbuf.Pixbuf.new_from_file(
            xdg.get_data_path('images', 'exailelogo.png')
        )
        self.set_logo(logo)

        import xl.version

        self.set_version(xl.version.__version__)

        # The user may have changed the theme since startup.
        theme = Gtk.Settings.get_default().props.gtk_theme_name
        xl.version.__external_versions__["GTK+ theme"] = theme

        comments = []
        for name, version in sorted(xl.version.__external_versions__.items()):
            comments.append('%s: %s' % (name, version))

        self.set_comments('\n'.join(comments))

    def on_response(self, *_):
        self.destroy()


@GtkTemplate('ui', 'shortcuts_dialog.ui')
class ShortcutsDialog(Gtk.Dialog):
    """
    Shows information about registered shortcuts

    TODO: someday upgrade to Gtk.ShortcutsWindow when we require 3.20 as
          a minimum GTK version. This would also enable automatically
          localized (translated) accelerator names.
    """

    # doesn't work if we don't set the treeview here too..
    shortcuts_treeview, shortcuts_model = GtkTemplate.Child.widgets(2)

    __gtype_name__ = 'ShortcutsDialog'

    def __init__(self, parent=None):
        Gtk.Dialog.__init__(self)
        self.init_template()

        self.set_transient_for(parent)

        for a in sorted(
            providers.get('mainwindow-accelerators'),
            key=lambda a: ('%04d' % (a.key)) + a.name,
        ):
            self.shortcuts_model.append(
                (Gtk.accelerator_get_label(a.key, a.mods), a.helptext.replace('_', ''))
            )

    @GtkTemplate.Callback
    def on_close_clicked(self, widget):
        self.destroy()


class MultiTextEntryDialog(Gtk.Dialog):
    """
    Exactly like a TextEntryDialog, except it can contain multiple
    labels/fields.

    Instead of using GetValue, use GetValues.  It will return a list with
    the contents of the fields. Each field must be filled out or the dialog
    will not close.
    """

    def __init__(self, parent, title):
        Gtk.Dialog.__init__(self, title=title, transient_for=parent)

        self.__entry_area = Gtk.Grid()
        self.__entry_area.set_row_spacing(3)
        self.__entry_area.set_column_spacing(3)
        self.__entry_area.set_border_width(3)
        self.vbox.pack_start(self.__entry_area, True, True, 0)

        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )

        self.fields = []

    def add_field(self, label):
        """
        Adds a field and corresponding label

        :param label: the label to display
        :returns: the newly created entry
        :rtype: :class:`Gtk.Entry`
        """
        line_number = len(self.fields)

        label = Gtk.Label(label=label)
        label.set_xalign(0)
        self.__entry_area.attach(label, 0, line_number, 1, 1)

        entry = Gtk.Entry()
        entry.set_width_chars(30)
        entry.set_icon_activatable(Gtk.EntryIconPosition.SECONDARY, False)
        entry.connect('activate', lambda *e: self.response(Gtk.ResponseType.OK))
        self.__entry_area.attach(entry, 1, line_number, 1, 1)
        label.show()
        entry.show()

        self.fields.append(entry)

        return entry

    def get_values(self):
        """
        Returns a list of the values from the added fields
        """
        return [a.get_text() for a in self.fields]

    def run(self):
        """
        Shows the dialog, runs, hides, and returns
        """
        self.show_all()

        while True:
            response = Gtk.Dialog.run(self)

            if response in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
                break

            if response == Gtk.ResponseType.OK:
                # Leave loop if all fields where filled
                if len(min([f.get_text() for f in self.fields])) > 0:
                    break

                # At least one field was not filled
                for field in self.fields:
                    if len(field.get_text()) > 0:
                        # Unset possible previous marks
                        field.set_icon_from_icon_name(
                            Gtk.EntryIconPosition.SECONDARY, None
                        )
                    else:
                        # Mark via warning
                        field.set_icon_from_icon_name(
                            Gtk.EntryIconPosition.SECONDARY, 'dialog-warning'
                        )
        self.hide()

        return response


class TextEntryDialog(Gtk.Dialog):
    """
    Shows a dialog with a single line of text
    """

    def __init__(
        self,
        message,
        title,
        default_text=None,
        parent=None,
        cancelbutton=None,
        okbutton=None,
    ):
        """
        Initializes the dialog
        """
        if not cancelbutton:
            cancelbutton = Gtk.STOCK_CANCEL
        if not okbutton:
            okbutton = Gtk.STOCK_OK
        Gtk.Dialog.__init__(
            self, title=title, transient_for=parent, destroy_with_parent=True
        )

        self.add_buttons(
            cancelbutton, Gtk.ResponseType.CANCEL, okbutton, Gtk.ResponseType.OK
        )

        label = Gtk.Label(label=message)
        label.set_xalign(0)
        self.vbox.set_border_width(5)

        main = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        main.set_spacing(3)
        main.set_border_width(5)
        self.vbox.pack_start(main, True, True, 0)

        main.pack_start(label, False, False, 0)

        self.entry = Gtk.Entry()
        self.entry.set_width_chars(35)
        if default_text:
            self.entry.set_text(default_text)
        main.pack_start(self.entry, False, False, 0)

        self.entry.connect('activate', lambda e: self.response(Gtk.ResponseType.OK))

    def get_value(self):
        """
        Returns the text value
        """
        return self.entry.get_text()

    def set_value(self, value):
        """
        Sets the value of the text
        """
        self.entry.set_text(value)

    def run(self):
        self.show_all()
        response = Gtk.Dialog.run(self)
        self.hide()
        return response


class URIOpenDialog(TextEntryDialog):
    """
    A dialog specialized for opening an URI
    """

    __gsignals__ = {
        'uri-selected': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (GObject.TYPE_PYOBJECT,),
            GObject.signal_accumulator_true_handled,
        )
    }

    def __init__(self, parent=None):
        """
        :param parent: a parent window for modal operation or None
        :type parent: :class:`Gtk.Window`
        """
        TextEntryDialog.__init__(
            self, message=_('Enter the URL to open'), title=_('Open URL'), parent=parent
        )

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        self.connect('response', self.on_response)

    def run(self):
        """
        Show the dialog and block until it's closed.

        The dialog will be automatically destroyed on user response.
        To obtain the entered URI, handle the "uri-selected" signal.
        """
        self.show()
        response = TextEntryDialog.run(self)
        self.emit('response', response)

    def show(self):
        """
        Show the dialog and return immediately.

        The dialog will be automatically destroyed on user response.
        To obtain the entered URI, handle the "uri-selected" signal.
        """
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        text = clipboard.wait_for_text()

        if text is not None:
            f = Gio.File.new_for_uri(text)
            if f.get_uri_scheme():
                self.set_value(text)

        TextEntryDialog.show_all(self)

    def do_uri_selected(self, uri):
        """
        Destroys the dialog
        """
        self.destroy()

    def on_response(self, dialog, response):
        """
        Notifies about the selected URI
        """
        self.hide()

        if response == Gtk.ResponseType.OK:
            self.emit('uri-selected', self.get_value())

        # self.destroy()

    '''
        dialog = dialogs.TextEntryDialog(_('Enter the URL to open'),
        _('Open URL'))
        dialog.set_transient_for(self.main.window)
        dialog.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        text = clipboard.wait_for_text()

        if text is not None:
            location = Gio.File.new_for_uri(text)

            if location.get_uri_scheme() is not None:
                dialog.set_value(text)

        result = dialog.run()
        dialog.hide()
        if result == Gtk.ResponseType.OK:
            url = dialog.get_value()
            self.open_uri(url, play=False)
    '''


class ListDialog(Gtk.Dialog):
    """
    Shows a dialog with a list of specified items

    Items must define a __str__ method, or be a string
    """

    def __init__(self, title, parent=None, multiple=False, write_only=False):
        """
        Initializes the dialog
        """
        Gtk.Dialog.__init__(self, title=title, transient_for=parent)

        self.vbox.set_border_width(5)
        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.model = Gtk.ListStore(object)
        self.list = Gtk.TreeView(model=self.model)
        self.list.set_headers_visible(False)
        self.list.connect(
            'row-activated', lambda *e: self.response(Gtk.ResponseType.OK)
        )
        scroll.add(self.list)
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        self.vbox.pack_start(scroll, True, True, 0)

        if write_only:
            self.add_buttons(Gtk.STOCK_OK, Gtk.ResponseType.OK)
        else:
            self.add_buttons(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OK,
                Gtk.ResponseType.OK,
            )

        self.selection = self.list.get_selection()

        if multiple:
            self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        else:
            self.selection.set_mode(Gtk.SelectionMode.SINGLE)

        text = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn()
        col.pack_start(text, True)

        col.set_cell_data_func(text, self.cell_data_func)
        self.list.append_column(col)
        self.list.set_model(self.model)
        self.resize(400, 240)

    def get_items(self):
        """
        Returns the selected items
        """
        items = []
        check = self.selection.get_selected_rows()
        if not check:
            return None
        (model, paths) = check

        for path in paths:
            iter = self.model.get_iter(path)
            item = self.model.get_value(iter, 0)
            items.append(item)

        return items

    def run(self):
        self.show_all()
        result = Gtk.Dialog.run(self)
        self.hide()
        return result

    def set_items(self, items):
        """
        Sets the items
        """
        for item in items:
            self.model.append([item])

    def cell_data_func(self, column, cell, model, iter, user_data):
        """
        Called when the tree needs a value for column 1
        """
        object = model.get_value(iter, 0)
        cell.set_property('text', str(object))


# TODO: combine this and list dialog


class ListBox:
    """
    Represents a list box
    """

    def __init__(self, widget, rows=None):
        """
        Initializes the widget
        """
        self.list = widget
        self.store = Gtk.ListStore(str)
        widget.set_headers_visible(False)
        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn('', cell, text=0)
        self.list.append_column(col)
        self.rows = rows
        if not rows:
            self.rows = []

        if rows:
            for row in rows:
                self.store.append([row])

        self.list.set_model(self.store)

    def connect(self, signal, func, data=None):
        """
        Connects a signal to the underlying treeview
        """
        self.list.connect(signal, func, data)

    def append(self, row):
        """
        Appends a row to the list
        """
        self.rows.append(row)
        self.set_rows(self.rows)

    def remove(self, row):
        """
        Removes a row
        """
        try:
            index = self.rows.index(row)
        except ValueError:
            return
        path = (index,)
        iter = self.store.get_iter(path)
        self.store.remove(iter)
        del self.rows[index]

    def set_rows(self, rows):
        """
        Sets the rows
        """
        self.rows = rows
        self.store = Gtk.ListStore(str)
        for row in rows:
            self.store.append([row])

        self.list.set_model(self.store)

    def get_selection(self):
        """
        Returns the selection
        """
        selection = self.list.get_selection()
        (model, iter) = selection.get_selected()
        if not iter:
            return None
        return model.get_value(iter, 0)


class FileOperationDialog(Gtk.FileChooserDialog):
    """
    An extension of the Gtk.FileChooserDialog that
    adds a collapsable panel to the bottom listing
    valid file extensions that the file can be
    saved in. (similar to the one in GIMP)
    """

    def __init__(
        self,
        title=None,
        parent=None,
        action=Gtk.FileChooserAction.OPEN,
        buttons=None,
        backend=None,
    ):
        """
        Standard __init__ of the Gtk.FileChooserDialog.
        Also sets up the expander and list for extensions
        """
        Gtk.FileChooserDialog.__init__(self, title, parent, action, buttons, backend)

        self.set_do_overwrite_confirmation(True)

        # Container for additional option widgets
        self.extras_box = Gtk.Box(spacing=3, orientation=Gtk.Orientation.VERTICAL)
        self.set_extra_widget(self.extras_box)
        self.extras_box.show()

        # Create the list that will hold the file type/extensions pair
        self.liststore = Gtk.ListStore(str, str)
        self.list = Gtk.TreeView(self.liststore)
        self.list.set_headers_visible(False)

        # Create the columns
        filetype_cell = Gtk.CellRendererText()
        extension_cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn()
        column.pack_start(filetype_cell, True)
        column.pack_start(extension_cell, False)
        column.set_attributes(filetype_cell, text=0)
        column.set_attributes(extension_cell, text=1)

        self.list.append_column(column)
        self.list.show_all()

        self.expander = Gtk.Expander.new(_('Select File Type (by Extension)'))
        self.expander.add(self.list)
        self.extras_box.pack_start(self.expander, False, False, 0)
        self.expander.show()

        selection = self.list.get_selection()
        selection.connect('changed', self.on_selection_changed)

    def on_selection_changed(self, selection):
        """
        When the user selects an extension the filename
        that is entered will have its extension changed
        to the selected extension
        """
        model, iter = selection.get_selected()
        (extension,) = model.get(iter, 1)
        filename = ""
        if self.get_filename():
            filename = os.path.basename(self.get_filename())
            filename, old_extension = os.path.splitext(filename)
            filename += '.' + extension
        else:
            filename = '*.' + extension
        self.set_current_name(filename)

    def add_extensions(self, extensions):
        """
        Adds extensions to the list

        @param extensions: a dictionary of extension:file type pairs
        i.e. { 'm3u':'M3U Playlist' }
        """
        for key, value in extensions.items():
            self.liststore.append((value, key))


class MediaOpenDialog(Gtk.FileChooserDialog):
    """
    A dialog for opening general media
    """

    __gsignals__ = {
        'uris-selected': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (GObject.TYPE_PYOBJECT,),
            GObject.signal_accumulator_true_handled,
        )
    }
    _last_location = None

    def __init__(self, parent=None):
        """
        :param parent: a parent window for modal operation or None
        :type parent: :class:`Gtk.Window`
        """
        Gtk.FileChooserDialog.__init__(
            self,
            title=_('Choose Media to Open'),
            parent=parent,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.set_local_only(False)
        self.set_select_multiple(True)

        supported_filter = Gtk.FileFilter()
        supported_filter.set_name(_('Supported Files'))
        audio_filter = Gtk.FileFilter()
        audio_filter.set_name(_('Music Files'))
        playlist_filter = Gtk.FileFilter()
        playlist_filter.set_name(_('Playlist Files'))
        all_filter = Gtk.FileFilter()
        all_filter.set_name(_('All Files'))
        all_filter.add_pattern('*')

        for extension in metadata.formats.keys():
            pattern = '*.%s' % extension
            supported_filter.add_pattern(pattern)
            audio_filter.add_pattern(pattern)

        playlist_file_extensions = (
            ext
            for p in providers.get('playlist-format-converter')
            for ext in p.file_extensions
        )

        for extension in playlist_file_extensions:
            pattern = '*.%s' % extension
            supported_filter.add_pattern(pattern)
            playlist_filter.add_pattern(pattern)

        self.add_filter(supported_filter)
        self.add_filter(audio_filter)
        self.add_filter(playlist_filter)
        self.add_filter(all_filter)

        self.connect('response', self.on_response)

    def run(self):
        """
        Override to take care of the response
        """
        if MediaOpenDialog._last_location is not None:
            self.set_current_folder_uri(MediaOpenDialog._last_location)

        response = Gtk.FileChooserDialog.run(self)
        self.emit('response', response)

    def show(self):
        """
        Override to restore last location
        """
        if MediaOpenDialog._last_location is not None:
            self.set_current_folder_uri(MediaOpenDialog._last_location)

        Gtk.FileChooserDialog.show(self)

    def do_uris_selected(self, uris):
        """
        Destroys the dialog
        """
        self.destroy()

    def on_response(self, dialog, response):
        """
        Notifies about selected URIs
        """
        self.hide()

        if response == Gtk.ResponseType.OK:
            MediaOpenDialog._last_location = self.get_current_folder_uri()
            self.emit('uris-selected', self.get_uris())

        # self.destroy()


class DirectoryOpenDialog(Gtk.FileChooserDialog):
    """
    A dialog specialized for opening directories
    """

    __gsignals__ = {
        'uris-selected': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (GObject.TYPE_PYOBJECT,),
            GObject.signal_accumulator_true_handled,
        )
    }
    _last_location = None

    def __init__(
        self, parent=None, title=_('Choose Directory to Open'), select_multiple=True
    ):
        Gtk.FileChooserDialog.__init__(
            self,
            title,
            parent=parent,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.set_action(Gtk.FileChooserAction.SELECT_FOLDER)
        self.set_local_only(False)
        self.set_select_multiple(select_multiple)

        self.connect('response', self.on_response)

    def run(self):
        """
        Override to take care of the response
        """
        if DirectoryOpenDialog._last_location is not None:
            self.set_current_folder_uri(DirectoryOpenDialog._last_location)

        Gtk.FileChooserDialog.run(self)

    def show(self):
        """
        Override to restore last location
        """
        if DirectoryOpenDialog._last_location is not None:
            self.set_current_folder_uri(DirectoryOpenDialog._last_location)

        Gtk.FileChooserDialog.show(self)

    def do_uris_selected(self, uris):
        """
        Destroys the dialog
        """
        self.destroy()

    def on_response(self, dialog, response):
        """
        Notifies about selected URIs
        """
        self.hide()

        if response == Gtk.ResponseType.OK:
            DirectoryOpenDialog._last_location = self.get_current_folder_uri()
            self.emit('uris-selected', self.get_uris())

        # self.destroy()


class PlaylistImportDialog(Gtk.FileChooserDialog):
    """
    A dialog for importing a playlist
    """

    __gsignals__ = {
        'playlists-selected': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (GObject.TYPE_PYOBJECT,),
            GObject.signal_accumulator_true_handled,
        )
    }
    _last_location = None

    def __init__(self, parent=None):
        """
        :param parent: a parent window for modal operation or None
        :type parent: :class:`Gtk.Window`
        """
        Gtk.FileChooserDialog.__init__(
            self,
            title=_('Import Playlist'),
            parent=parent,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.set_local_only(False)
        self.set_select_multiple(True)

        playlist_filter = Gtk.FileFilter()
        playlist_filter.set_name(_('Playlist Files'))
        all_filter = Gtk.FileFilter()
        all_filter.set_name(_('All Files'))
        all_filter.add_pattern('*')

        playlist_file_extensions = (
            ext
            for p in providers.get('playlist-format-converter')
            for ext in p.file_extensions
        )

        for extension in playlist_file_extensions:
            pattern = '*.%s' % extension
            playlist_filter.add_pattern(pattern)

        self.add_filter(playlist_filter)
        self.add_filter(all_filter)

        self.connect('response', self.on_response)

    def run(self):
        """
        Override to take care of the response
        """
        if PlaylistImportDialog._last_location is not None:
            self.set_current_folder_uri(PlaylistImportDialog._last_location)

        response = Gtk.FileChooserDialog.run(self)
        self.emit('response', response)

    def show(self):
        """
        Override to restore last location
        """
        if PlaylistImportDialog._last_location is not None:
            self.set_current_folder_uri(PlaylistImportDialog._last_location)

        Gtk.FileChooserDialog.show(self)

    def do_playlist_selected(self, uris):
        """
        Destroys the dialog
        """
        self.destroy()

    def on_response(self, dialog, response):
        """
        Notifies about selected URIs
        """
        self.hide()

        if response == Gtk.ResponseType.OK:
            PlaylistImportDialog._last_location = self.get_current_folder_uri()

            playlists = []
            for uri in self.get_uris():
                try:
                    playlists.append(import_playlist(uri))
                except InvalidPlaylistTypeError as e:
                    error(
                        self.get_transient_for(), 'Invalid playlist "%s": %s' % (uri, e)
                    )
                    self.destroy()
                    return
                except Exception as e:
                    error(
                        self.get_transient_for(),
                        'Invalid playlist "%s": (internal error): %s' % (uri, e),
                    )
                    self.destroy()
                    return

            self.emit('playlists-selected', playlists)

        # self.destroy()


class PlaylistExportDialog(FileOperationDialog):
    """
    A dialog specialized for playlist export
    """

    __gsignals__ = {
        'message': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (Gtk.MessageType, GObject.TYPE_STRING),
            GObject.signal_accumulator_true_handled,
        )
    }

    def __init__(self, playlist, parent=None):
        """
        :param playlist: the playlist to export
        :type playlist: :class:`xl.playlist.Playlist`
        :param parent: a parent window for modal operation or None
        :type parent: :class:`Gtk.Window`
        """
        FileOperationDialog.__init__(
            self,
            title=_('Export Current Playlist'),
            parent=parent,
            action=Gtk.FileChooserAction.SAVE,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_SAVE,
                Gtk.ResponseType.OK,
            ),
        )

        self.set_current_folder_uri(
            settings.get_option('gui/playlist_export_dir')
            or GLib.filename_to_uri(xdg.homedir, None)
        )

        self.set_local_only(False)

        self.relative_checkbox = Gtk.CheckButton(_('Use relative paths to tracks'))
        self.relative_checkbox.set_active(True)
        self.extras_box.pack_start(self.relative_checkbox, False, False, 3)
        self.relative_checkbox.show()

        self.playlist = playlist

        extensions = {}

        for provider in providers.get('playlist-format-converter'):
            extensions[provider.name] = provider.title

        self.add_extensions(extensions)
        self.set_current_name('%s.m3u' % playlist.name)

        self.connect('response', self.on_response)

    def run(self):
        """
        Override to take care of the response
        """
        response = FileOperationDialog.run(self)
        self.emit('response', response)

    def do_message(self, message_type, message):
        """
        Displays simple dialogs on messages
        """
        if message_type == Gtk.MessageType.INFO:
            info(self.get_transient_for(), markup=message)
        elif message_type == Gtk.MessageType.ERROR:
            error(self.get_transient_for(), markup=message)

    def on_response(self, dialog, response):
        """
        Exports the playlist if requested
        """
        self.hide()

        if response == Gtk.ResponseType.OK:
            gfile = self.get_file()
            settings.set_option('gui/playlist_export_dir', gfile.get_parent().get_uri())

            path = gfile.get_uri()
            if not is_valid_playlist(path):
                path = '%s.m3u' % path

            options = PlaylistExportOptions(
                relative=self.relative_checkbox.get_active()
            )

            try:
                export_playlist(self.playlist, path, options)
            except InvalidPlaylistTypeError as e:
                self.emit('message', Gtk.MessageType.ERROR, str(e))
            else:
                self.emit(
                    'message',
                    Gtk.MessageType.INFO,
                    _('Playlist saved as <b>%s</b>.') % path,
                )

        # self.destroy()


class ConfirmCloseDialog(Gtk.MessageDialog):
    """
    Shows the dialog to confirm closing of the playlist
    """

    def __init__(self, document_name):
        """
        Initializes the dialog
        """
        Gtk.MessageDialog.__init__(self, type=Gtk.MessageType.WARNING)

        self.set_title(_('Close %s' % document_name))
        self.set_markup(_('<b>Save changes to %s before closing?</b>') % document_name)
        self.format_secondary_text(
            _('Your changes will be lost if you don\'t save them')
        )

        self.add_buttons(
            _('Close Without Saving'),
            100,
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_SAVE,
            110,
        )

    def run(self):
        self.show_all()
        response = Gtk.Dialog.run(self)
        self.hide()
        return response


class MessageBar(Gtk.InfoBar):
    type_map = {
        Gtk.MessageType.INFO: 'dialog-information',
        Gtk.MessageType.QUESTION: 'dialog-question',
        Gtk.MessageType.WARNING: 'dialog-warning',
        Gtk.MessageType.ERROR: 'dialog-error',
    }
    buttons_map = {
        Gtk.ButtonsType.OK: [(Gtk.STOCK_OK, Gtk.ResponseType.OK)],
        Gtk.ButtonsType.CLOSE: [(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)],
        Gtk.ButtonsType.CANCEL: [(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)],
        Gtk.ButtonsType.YES_NO: [
            (Gtk.STOCK_NO, Gtk.ResponseType.NO),
            (Gtk.STOCK_YES, Gtk.ResponseType.YES),
        ],
        Gtk.ButtonsType.OK_CANCEL: [
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL),
            (Gtk.STOCK_OK, Gtk.ResponseType.OK),
        ],
    }

    def __init__(
        self,
        parent=None,
        type=Gtk.MessageType.INFO,
        buttons=Gtk.ButtonsType.NONE,
        text=None,
    ):
        """
        Report important messages to the user

        :param parent: the parent container box
        :type parent: :class:`Gtk.Box`
        :param type: the type of message: Gtk.MessageType.INFO,
            Gtk.MessageType.WARNING, Gtk.MessageType.QUESTION or
            Gtk.MessageType.ERROR.
        :param buttons: the predefined set of buttons to
            use: Gtk.ButtonsType.NONE, Gtk.ButtonsType.OK,
            Gtk.ButtonsType.CLOSE, Gtk.ButtonsType.CANCEL,
            Gtk.ButtonsType.YES_NO, Gtk.ButtonsType.OK_CANCEL
        :param text: a string containing the message
            text or None
        """
        if parent is not None and not isinstance(parent, Gtk.Box):
            raise TypeError('Parent needs to be of type Gtk.Box')

        Gtk.InfoBar.__init__(self)
        self.set_no_show_all(True)

        if parent is not None:
            parent.add(self)
            parent.reorder_child(self, 0)
            parent.child_set_property(self, 'expand', False)

        self.image = Gtk.Image()
        self.set_message_type(type)

        self.primary_text = Gtk.Label(label=text)
        self.primary_text.set_property('xalign', 0)
        self.primary_text.set_line_wrap(True)
        self.secondary_text = Gtk.Label()
        self.secondary_text.set_property('xalign', 0)
        self.secondary_text.set_line_wrap(True)
        self.secondary_text.set_no_show_all(True)
        self.secondary_text.set_selectable(True)

        self.message_area = Gtk.Box(spacing=12, orientation=Gtk.Orientation.VERTICAL)
        self.message_area.pack_start(self.primary_text, False, False, 0)
        self.message_area.pack_start(self.secondary_text, False, False, 0)

        box = Gtk.Box(spacing=6)
        box.pack_start(self.image, False, True, 0)
        box.pack_start(self.message_area, True, True, 0)

        content_area = self.get_content_area()
        content_area.add(box)
        content_area.show_all()

        self.action_area = self.get_action_area()
        self.action_area.set_property('layout-style', Gtk.ButtonBoxStyle.START)

        if buttons != Gtk.ButtonsType.NONE:
            for text, response in self.buttons_map[buttons]:
                self.add_button(text, response)

        self.primary_text_attributes = Pango.AttrList()
        # TODO: GI: Pango attr
        # self.primary_text_attributes.insert(
        #    Pango.AttrWeight(Pango.Weight.NORMAL, 0, -1))
        # self.primary_text_attributes.insert(
        #    Pango.AttrScale(Pango.SCALE_MEDIUM, 0, -1))'''

        self.primary_text_emphasized_attributes = Pango.AttrList()
        # TODO: GI: Pango attr
        # self.primary_text_emphasized_attributes.insert(
        #    Pango.AttrWeight(Pango.Weight.BOLD, 0, -1))
        # self.primary_text_emphasized_attributes.insert(
        #    Pango.AttrScale(Pango.SCALE_LARGE, 0, -1))'''

        self.connect('response', self.on_response)

        # Workaround for https://bugzilla.gnome.org/show_bug.cgi?id=710888
        # -> From pitivi: https://phabricator.freedesktop.org/D1103#34aa2703
        def _make_sure_revealer_does_nothing(widget):
            if not isinstance(widget, Gtk.Revealer):
                return
            widget.set_transition_type(Gtk.RevealerTransitionType.NONE)

        self.forall(_make_sure_revealer_does_nothing)

    def set_text(self, text):
        """
        Sets the primary text of the message bar

        :param markup: a regular text string
        :type markup: string
        """
        self.primary_text.set_text(text)

    def set_markup(self, markup):
        """
        Sets the primary markup text of the message bar

        :param markup: a markup string
        :type markup: string
        """
        self.primary_text.set_markup(markup)

    def set_secondary_text(self, text):
        """
        Sets the secondary text to the text
        specified by message_format

        :param text: The text to be displayed
            as the secondary text or None.
        :type text: string
        """
        if text is None:
            self.secondary_text.hide()
            self.primary_text.set_attributes(self.primary_text_attributes)
        else:
            self.secondary_text.set_text(text)
            self.secondary_text.show()
            self.primary_text.set_attributes(self.primary_text_emphasized_attributes)

    def set_secondary_markup(self, markup):
        """
        Sets the secondary text to the markup text
        specified by text.

        :param text: A string containing the
            pango markup to use as secondary text.
        :type text: string
        """
        if markup is None:
            self.secondary_text.hide()
            self.primary_text.set_attributes(self.primary_text_attributes)
        else:
            self.secondary_text.set_markup(markup)
            self.secondary_text.show()
            self.primary_text.set_attributes(self.primary_text_emphasized_attributes)

    def set_image(self, image):
        """
        Sets the contained image to the :class:`Gtk.Widget`
        specified by image.

        :param image: the image widget
        :type image: :class:`Gtk.Widget`
        """
        box = self.image.get_parent()
        box.remove(self.image)
        self.image = image
        box.pack_start(self.image, False, True, 0)
        box.reorder_child(self.image, 0)
        self.image.show()

    def get_image(self):
        """
        Gets the contained image
        """
        return self.image

    def add_button(self, button_text, response_id):
        """
        Overrides :class:`Gtk.InfoBar` to prepend
        instead of append to the action area

        :param button_text: text of button, or stock ID
        :type button_text: string
        :param response_id: response ID for the button
        :type response_id: int
        """
        button = Gtk.InfoBar.add_button(self, button_text, response_id)
        self.action_area.reorder_child(button, 0)

        return button

    def clear_buttons(self):
        """
        Removes all buttons currently
        placed in the action area
        """
        for button in self.action_area:
            self.action_area.remove(button)

    def set_message_type(self, type):
        """
        Sets the message type of the message area.

        :param type: the type of message: Gtk.MessageType.INFO,
            Gtk.MessageType.WARNING, Gtk.MessageType.QUESTION or
            Gtk.MessageType.ERROR.
        """
        if type != Gtk.MessageType.OTHER:
            self.image.set_from_icon_name(self.type_map[type], Gtk.IconSize.DIALOG)

        Gtk.InfoBar.set_message_type(self, type)

    def get_message_area(self):
        """
        Retrieves the message area
        """
        return self.message_area

    def _show_message(
        self, message_type, text, secondary_text, markup, secondary_markup, timeout
    ):
        """
        Helper for the various `show_*` methods. See `show_info` for
        documentation on the parameters.
        """
        if text is markup is None:
            raise ValueError("text or markup must be specified")

        self.set_message_type(message_type)
        if markup is None:
            self.set_text(text)
        else:
            self.set_markup(markup)
        if secondary_markup is None:
            self.set_secondary_text(secondary_text)
        else:
            self.set_secondary_markup(secondary_markup)
        self.show()

        if timeout > 0:
            GLib.timeout_add_seconds(timeout, self.hide)

    def show_info(
        self,
        text=None,
        secondary_text=None,
        markup=None,
        secondary_markup=None,
        timeout=5,
    ):
        """
        Convenience method which sets all
        required flags for an info message

        :param text: the message to display
        :type text: string
        :param secondary_text: additional information
        :type secondary_text: string
        :param markup: the message to display, in Pango markup format
            (overrides `text`)
        :type markup: string
        :param secondary_markup: additional information, in Pango markup
            format (overrides `secondary_text`)
        :type secondary_markup: string
        :param timeout: after how many seconds the
            message should be hidden automatically,
            use 0 to disable this behavior
        :type timeout: int
        """
        self._show_message(
            Gtk.MessageType.INFO,
            text,
            secondary_text,
            markup,
            secondary_markup,
            timeout,
        )

    def show_question(
        self, text=None, secondary_text=None, markup=None, secondary_markup=None
    ):
        """
        Convenience method which sets all
        required flags for a question message

        :param text: the message to display
        :param secondary_text: additional information
        :param markup: the message to display, in Pango markup format
        :param secondary_markup: additional information, in Pango markup format
        """
        self._show_message(
            Gtk.MessageType.QUESTION, text, secondary_text, markup, secondary_markup, 0
        )

    def show_warning(
        self, text=None, secondary_text=None, markup=None, secondary_markup=None
    ):
        """
        Convenience method which sets all
        required flags for a warning message

        :param text: the message to display
        :param secondary_text: additional information
        :param markup: the message to display, in Pango markup format
        :param secondary_markup: additional information, in Pango markup format
        """
        self._show_message(
            Gtk.MessageType.WARNING, text, secondary_text, markup, secondary_markup, 0
        )

    def show_error(
        self, text=None, secondary_text=None, markup=None, secondary_markup=None
    ):
        """
        Convenience method which sets all
        required flags for a warning message

        :param text: the message to display
        :param secondary_text: additional information
        :param markup: the message to display, in Pango markup format
        :param secondary_markup: additional information, in Pango markup format
        """
        self._show_message(
            Gtk.MessageType.ERROR, text, secondary_text, markup, secondary_markup, 0
        )

    def on_response(self, widget, response):
        """
        Handles the response for closing
        """
        if response == Gtk.ResponseType.CLOSE:
            self.hide()


#
# Message ID's used by the XMessageDialog
#

XRESPONSE_YES = Gtk.ResponseType.YES
XRESPONSE_YES_ALL = 8000
XRESPONSE_NO = Gtk.ResponseType.NO
XRESPONSE_NO_ALL = 8001
XRESPONSE_CANCEL = Gtk.ResponseType.CANCEL


class XMessageDialog(Gtk.Dialog):
    '''Used to show a custom message dialog with custom buttons'''

    def __init__(
        self,
        title,
        text,
        parent=None,
        show_yes=True,
        show_yes_all=True,
        show_no=True,
        show_no_all=True,
        show_cancel=True,
    ):

        Gtk.Dialog.__init__(self, title=title, transient_for=parent)

        #
        # TODO: Make these buttons a bit prettier
        #

        if show_yes:
            self.add_button(Gtk.STOCK_YES, XRESPONSE_YES)
            self.set_default_response(XRESPONSE_YES)

        if show_yes_all:
            self.add_button(_('Yes to all'), XRESPONSE_YES_ALL)
            self.set_default_response(XRESPONSE_YES_ALL)

        if show_no:
            self.add_button(Gtk.STOCK_NO, XRESPONSE_NO)
            self.set_default_response(XRESPONSE_NO)

        if show_no_all:
            self.add_button(_('No to all'), XRESPONSE_NO_ALL)
            self.set_default_response(XRESPONSE_NO_ALL)

        if show_cancel:
            self.add_button(Gtk.STOCK_CANCEL, XRESPONSE_CANCEL)
            self.set_default_response(XRESPONSE_CANCEL)

        vbox = self.get_content_area()
        self._label = Gtk.Label()
        self._label.set_use_markup(True)
        self._label.set_markup(text)
        vbox.pack_start(self._label, True, True, 0)


class FileCopyDialog(Gtk.Dialog):
    """
    Used to copy a list of files to a single destination directory

    Usage:
        dialog = FileCopyDialog( [file_uri,..], destination_uri, text, parent)
        dialog.do_copy()

    Do not use run() on this dialog!
    """

    class CopyThread(Thread):
        def __init__(self, source, dest, callback_finish_single_copy, copy_flags):
            Thread.__init__(self, name='CopyThread')
            self.__source = source
            self.__dest = dest
            self.__cb_single_copy = callback_finish_single_copy
            self.__copy_flags = copy_flags
            self.__cancel = Gio.Cancellable()
            self.start()

        def run(self):
            try:
                result = self.__source.copy(
                    self.__dest, flags=self.__copy_flags, cancellable=self.__cancel
                )
                GLib.idle_add(self.__cb_single_copy, self.__source, result, None)
            except GLib.Error as err:
                GLib.idle_add(self.__cb_single_copy, self.__source, False, err)

        def cancel_copy(self):
            self.__cancel.cancel()

    def __init__(
        self,
        file_uris,
        destination_uri,
        title,
        text=_("Saved %(count)s of %(total)s."),
        parent=None,
    ):

        self.file_uris = file_uris
        self.destination_uri = destination_uri
        self.is_copying = False

        Gtk.Dialog.__init__(self, title=title, transient_for=parent)

        self.parent = parent
        self.count = 0
        self.total = len(file_uris)
        self.text = text
        self.overwrite_response = None

        # self.set_modal(True)
        # self.set_decorated(False)
        self.set_resizable(False)
        # self.set_focus_on_map(False)

        vbox = self.get_content_area()

        vbox.set_spacing(12)
        vbox.set_border_width(12)

        self._label = Gtk.Label()
        self._label.set_use_markup(True)
        self._label.set_markup(self.text % {'count': 0, 'total': self.total})
        vbox.pack_start(self._label, True, True, 0)

        self._progress = Gtk.ProgressBar()
        self._progress.set_size_request(300, -1)
        vbox.pack_start(self._progress, True, True, 0)

        self.show_all()

        # TODO: Make dialog cancelable
        # self.cancel_button.connect('activate', lambda *e: self.cancel.cancel() )

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

    def do_copy(self):
        logger.info("Copy started.")
        self._start_next_copy()
        self.show_all()
        self.connect('response', self._on_response)

    def run(self):
        raise NotImplementedError("Don't use this")

    def _on_response(self, widget, response):
        logger.info("Copy complete.")
        self.destroy()

    def _step(self):
        '''Steps the progress bar'''
        self.count += 1
        self._progress.set_fraction(clamp(self.count / float(self.total), 0, 1))
        self._label.set_markup(self.text % {'count': self.count, 'total': self.total})

    def _start_next_copy(self, overwrite=False):

        if self.count == len(self.file_uris):
            self.response(Gtk.ResponseType.OK)
            return

        flags = Gio.FileCopyFlags.NONE

        src_uri = self.file_uris[self.count]
        dst_uri = self.destination_uri + '/' + src_uri.split('/')[-1]

        self.source = Gio.File.new_for_uri(src_uri)
        self.destination = Gio.File.new_for_uri(dst_uri)

        if not overwrite:
            if self.destination.query_exists(None):
                if self.overwrite_response == XRESPONSE_YES_ALL:
                    overwrite = True

                elif (
                    self.overwrite_response == XRESPONSE_NO_ALL
                    or self.overwrite_response == XRESPONSE_NO
                ):

                    # only deny the overwrite once..
                    if self.overwrite_response == XRESPONSE_NO:
                        self.overwrite_response = None

                    logging.info("NoOverwrite: %s" % self.destination.get_uri())
                    self._step()
                    GLib.idle_add(self._start_next_copy)  # don't recurse
                    return
                else:
                    self._query_overwrite()
                    return

        if overwrite:
            flags = Gio.FileCopyFlags.OVERWRITE
            try:
                # Gio.FileCopyFlags.OVERWRITE doesn't actually work
                logging.info("DeleteDest : %s" % self.destination.get_uri())
                self.destination.delete()
            except GLib.Error:
                pass

        logging.info("CopySource : %s" % self.source.get_uri())
        logging.info("CopyDest   : %s" % self.destination.get_uri())

        # TODO g_file_copy_async() isn't introspectable
        # see https://github.com/exaile/exaile/issues/198 for details
        # self.source.copy_async( self.destination, self._finish_single_copy_async, flags=flags, cancellable=self.cancel )

        self.cpthr = self.CopyThread(
            self.source, self.destination, self._finish_single_copy, flags
        )

    def _finish_single_copy(self, source, success, error):
        if error:
            self._on_error(
                _("Error occurred while copying %s: %s")
                % (
                    GLib.markup_escape_text(self.source.get_uri()),
                    GLib.markup_escape_text(str(error)),
                )
            )
        if success:
            self._step()
            self._start_next_copy()

    def _finish_single_copy_async(self, source, async_result):

        try:
            if source.copy_finish(async_result):
                self._step()
                self._start_next_copy()
        except GLib.Error as e:
            self._on_error(
                _("Error occurred while copying %s: %s")
                % (
                    GLib.markup_escape_text(self.source.get_uri()),
                    GLib.markup_escape_text(str(e)),
                )
            )

    def _query_overwrite(self):

        self.hide()

        text = _('File exists, overwrite %s ?') % GLib.markup_escape_text(
            self.destination.get_uri()
        )
        dialog = XMessageDialog(self.parent, text)
        dialog.connect('response', self._on_query_overwrite_response, dialog)
        dialog.show_all()
        dialog.grab_focus()
        self.query_dialog = dialog

    def _on_query_overwrite_response(self, widget, response, dialog):
        dialog.destroy()
        self.overwrite_response = response

        if response == Gtk.ResponseType.CANCEL:
            self.response(response)
        else:
            if response == XRESPONSE_NO or response == XRESPONSE_NO_ALL:
                overwrite = False
            else:
                overwrite = True

            self.show_all()
            self._start_next_copy(overwrite)

    def _on_error(self, message):

        self.hide()

        dialog = Gtk.MessageDialog(
            buttons=Gtk.ButtonsType.CLOSE,
            message_type=Gtk.MessageType.ERROR,
            modal=True,
            text=message,
            transient_for=self.parent,
        )
        dialog.set_markup(message)
        dialog.connect('response', self._on_error_response, dialog)
        dialog.show()
        dialog.grab_focus()
        self.error_dialog = dialog

    def _on_error_response(self, widget, response, dialog):
        self.response(Gtk.ResponseType.CANCEL)
        dialog.destroy()


def ask_for_playlist_name(parent, playlist_manager, name=None):
    """
    Returns a user-selected name that is not already used
        in the specified playlist manager

    :param name: A default name to show to the user
    Returns None if the user hits cancel
    """

    while True:

        dialog = TextEntryDialog(
            _('Playlist name:'),
            _('Add new playlist...'),
            name,
            parent=parent,
            okbutton=Gtk.STOCK_ADD,
        )

        result = dialog.run()
        if result != Gtk.ResponseType.OK:
            return None

        name = dialog.get_value()

        if name == '':
            error(parent, _("You did not enter a name for your playlist"))
        elif playlist_manager.has_playlist_name(name):
            # name is already in use
            error(parent, _("The playlist name you entered is already in use."))
        else:
            return name


def save(
    parent, output_fname, output_setting=None, extensions=None, title=_("Save As")
):
    """
    A 'save' dialog utility function, which can be used to easily
    remember the last location the user saved something.

    :param parent:          Parent window
    :param output_fname:    Output filename
    :param output_setting:  Setting to store the last 'output directory' saved at
    :param extensions:      Valid output extensions. Dict { '.m3u': 'Description', .. }
    :param title:           Title of dialog

    :returns: None if user cancels, chosen URI otherwise
    """

    uri = None

    dialog = FileOperationDialog(
        title,
        parent,
        Gtk.FileChooserAction.SAVE,
        (
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_SAVE,
            Gtk.ResponseType.ACCEPT,
        ),
    )

    if extensions is not None:
        dialog.add_extensions(extensions)

    dialog.set_current_name(output_fname)

    if output_setting:
        output_dir = settings.get_option(output_setting)
        if output_dir:
            dialog.set_current_folder_uri(output_dir)

    if dialog.run() == Gtk.ResponseType.ACCEPT:
        uri = dialog.get_uri()

        settings.set_option(output_setting, dialog.get_current_folder_uri())

    dialog.destroy()

    return uri


def export_playlist_dialog(playlist, parent=None):
    '''Exports the playlist to a user-specified path'''
    if playlist is not None:
        dialog = PlaylistExportDialog(playlist, parent)
        dialog.show()


def export_playlist_files(playlist, parent=None):
    '''Exports the playlist files to a user-specified URI'''

    if playlist is None:
        return

    def _on_uri(uri):
        if hasattr(playlist, 'get_playlist'):
            pl = playlist.get_playlist()
        else:
            pl = playlist
        pl_files = [track.get_loc_for_io() for track in pl]
        dialog = FileCopyDialog(
            pl_files, uri, _('Exporting %s') % playlist.name, parent=parent
        )
        dialog.do_copy()

    dialog = DirectoryOpenDialog(
        title=_('Choose directory to export files to'), parent=parent
    )
    dialog.set_select_multiple(False)
    dialog.connect('uris-selected', lambda widget, uris: _on_uri(uris[0]))
    dialog.run()
    dialog.destroy()



import logging
from xl import event, settings
from . import exaile_parser
from .server import DaapServer
from . import daapserverprefs

logger = logging.getLogger(__file__)


# todo support multiple connections?
class CollectionWrapper:
    '''Class to wrap Exaile's collection to make it spydaap compatible'''

    class TrackWrapper:
        '''Wrap a single track for spydaap'''

        def __init__(self, id, track):
            self.track = track
            self.id = id
            self.parser = exaile_parser.ExaileParser()
            self.daap = None

        def get_dmap_raw(self):
            if self.daap is None:
                do = self.parser.parse(self.track)[0]
                if do is not None:
                    self.daap = b''.join([d.encode() for d in do])
                else:
                    self.daap = b''
            return self.daap

        def get_original_filename(self):
            return self.track.get_local_path()

    def __init__(self, collection):
        self.collection = collection
        self.map = []

    def __iter__(self):
        i = 0
        self.map = []
        for t in self.collection:
            self.map.append(self.TrackWrapper(i, t))
            yield self.map[i]
            i += 1

    def get_item_by_id(self, id):
        return self.map[int(id)]

    def __getitem__(self, idx):
        return self.map[idx]

    def __len__(self):
        return len(self.collection)


class DaapServerPlugin:
    __exaile = None
    __daapserver = None

    def on_gui_loaded(self):
        event.add_callback(self.__on_settings_changed, 'plugin_daapserver_option_set')

        port = int(settings.get_option('plugin/daapserver/port', 3689))
        name = settings.get_option('plugin/daapserver/name', 'Exaile Share')
        host = settings.get_option('plugin/daapserver/host', '0.0.0.0')

        self.__daapserver = DaapServer(
            CollectionWrapper(self.__exaile.collection), port=port, name=name, host=host
        )
        if settings.get_option('plugin/daapserver/enabled', True):
            self.__daapserver.start()

    def enable(self, exaile):
        self.__exaile = exaile

    def teardown(self, exaile):
        self.__daapserver.stop_server()

    def disable(self, exaile):
        self.teardown(exaile)
        self.__daapserver = None

    def get_preferences_pane(self):
        return daapserverprefs

    def __on_settings_changed(self, event, setting, option):
        if self.__daapserver is None:
            logger.error('Option set on uninitialized plugin. This is wrong.')
        if option == 'plugin/daapserver/name':
            self.__daapserver.set(name=settings.get_option(option, 'Exaile Share'))
        if option == 'plugin/daapserver/port':
            self.__daapserver.set(port=int(settings.get_option(option, 3689)))
        if option == 'plugin/daapserver/host':
            self.__daapserver.set(host=settings.get_option(option, '0.0.0.0'))
        if option == 'plugin/daapserver/enabled':
            enabled = setting.get_option(option, True)
            if enabled:
                if not self.__daapserver.start():
                    logger.error('failed to start DAAP Server.')
            else:
                self.__daapserver.stop_server()


plugin_class = DaapServerPlugin



# Copyright (C) 2009 Aren Olson (for the old_get_track_key function)
# Copyright (C) 2018 Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

__all__ = ['migrate']

import logging
import os

import xl.collection
import xl.covers

logger = logging.getLogger(__name__)


def old_get_track_key(track):
    """
    Get the db mapping key for a track
    """
    album = track.get_tag_raw("album", join=True)
    compilation = track.get_tag_raw("__compilation")

    if compilation:
        value = track.get_tag_raw('albumartist')
        if value:
            tag = 'albumartist'
        else:
            tag = 'compilation'
            value = compilation
    elif album:
        tag = 'album'
        value = album
    else:
        # no album info, cant store it
        return None
    return (tag, tuple(value))


def migrate():
    """Migrate covers.db version 1 to 2 (Exaile 4.0)."""

    man = xl.covers.MANAGER
    if man.db.get('version', 1) != 1:
        return
    logger.info("Upgrading covers.db to version 2")

    valid_cachefiles = set()

    old_db = man.db
    new_db = {'version': 2}
    for coll in xl.collection.COLLECTIONS:
        for tr in coll.tracks.values():
            key = old_get_track_key(tr._track)
            value = old_db.get(key)
            if value:
                new_key = man._get_track_key(tr)
                new_db[new_key] = value
                if value.startswith('cache:'):
                    valid_cachefiles.add(value[6:])
    man.db = new_db
    man.save()

    cachedir = os.path.join(man.location, 'cache')
    for cachefile in frozenset(os.listdir(cachedir)) - valid_cachefiles:
        os.remove(os.path.join(cachedir, cachefile))



# Copyright (C) 2014 Rocco Aliberti
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

try:
    import lxml.html
except ImportError:
    lxml = None

import re

from xl.lyrics import LyricSearchMethod, LyricsNotFoundException
from xl import common, providers


def enable(exaile):
    """
    Enables the lyrics mania plugin that fetches track lyrics
    from lyricsmania.com
    """
    if lxml:
        providers.register('lyrics', LyricsMania(exaile))
    else:
        raise NotImplementedError('LXML is not available.')
        return False


def disable(exaile):
    providers.unregister('lyrics', providers.get_provider('lyrics', 'lyricsmania'))


class LyricsMania(LyricSearchMethod):

    name = "lyricsmania"
    display_name = "Lyrics Mania"

    def __init__(self, exaile):
        self.user_agent = exaile.get_user_agent_string('lyricsmania')

    def find_lyrics(self, track):
        try:
            (artist, title) = (
                track.get_tag_raw('artist')[0],
                track.get_tag_raw('title')[0],
            )
        except TypeError:
            raise LyricsNotFoundException

        if not artist or not title:
            raise LyricsNotFoundException

        artist = artist.replace(' ', '_').replace('\'', '').lower()
        title = title.replace(' ', '_').replace('\'', '').lower()

        url = 'https://www.lyricsmania.com/%s_lyrics_%s.html' % (title, artist)

        try:
            html = common.get_url_contents(url, self.user_agent)
        except Exception:
            raise LyricsNotFoundException

        try:
            lyrics_html = lxml.html.fromstring(html)
        except lxml.etree.XMLSyntaxError:
            raise LyricsNotFoundException

        try:
            lyrics_body = lyrics_html.find_class('lyrics-body')[0]
            lyrics_body.remove(lyrics_body.get_element_by_id('video-musictory'))
            lyrics = re.sub(r'^\s+Lyrics to .+', '', lyrics_body.text_content())
            lyrics = lyrics.replace('\t', '')
            lyrics = self.remove_script(lyrics)
            lyrics = self.remove_html_tags(lyrics)
        except Exception:
            raise LyricsNotFoundException

        # We end up with unicode in some systems, str (bytes) in others;
        # no idea why and which one is correct.
        if isinstance(lyrics, bytes):
            lyrics = lyrics.decode('utf-8', errors='replace')
        return (lyrics, self.name, url)



# Keybinder-based multimedia keys support
# Copyright (C) 2015  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import gi

from xlgui import guiutil
from xl.nls import gettext as _

gi.require_version('Keybinder', '3.0')

from gi.repository import Keybinder


LOGGER = logging.getLogger(__name__)

KEYS = [
    'XF86AudioPlay',
    'XF86AudioStop',
    'XF86AudioPrev',
    'XF86AudioNext',
    'XF86AudioPause',
    'XF86AudioMedia',
    'XF86AudioRewind',
]


class KeybinderPlugin:
    def __init__(self):
        self.__exaile = None

    def enable(self, exaile):
        broken = False
        if hasattr(Keybinder, 'supported'):
            # introduced in Keybinder-3.0 0.3.2, see
            # https://github.com/kupferlauncher/keybinder/blob/master/NEWS
            if not Keybinder.supported():
                broken = True
        elif not guiutil.platform_is_x11():
            broken = True
        if broken:
            raise Exception(
                _(
                    'Keybinder is not supported on this platform! '
                    'It is only supported on X servers.'
                )
            )
        self.__exaile = exaile

    def on_exaile_loaded(self):
        if not self.__exaile:
            return  # Plugin has been disabled in the meantime
        Keybinder.init()
        for k in KEYS:
            if not Keybinder.bind(k, on_media_key, self.__exaile):
                LOGGER.warning("Failed to set key binding using Keybinder.")
                self.__exaile.plugins.disable_plugin(__name__)
                return

    def teardown(self, exaile):
        for k in KEYS:
            Keybinder.unbind(k)

    def disable(self, exaile):
        self.teardown(exaile)
        self.__exaile = None


plugin_class = KeybinderPlugin


def start_stop_playback(PLAYER, QUEUE):
    '''Toggles pause if playing/paused, starts playback if stopped'''
    if PLAYER.is_paused() or PLAYER.is_playing():
        PLAYER.toggle_pause()
    else:
        QUEUE.play(track=QUEUE.get_current())


def on_media_key(key, exaile):
    from xl.player import PLAYER, QUEUE

    {
        'XF86AudioPlay': lambda: start_stop_playback(PLAYER, QUEUE),
        'XF86AudioStop': PLAYER.stop,
        'XF86AudioPrev': QUEUE.prev,
        'XF86AudioNext': QUEUE.next,
        'XF86AudioPause': PLAYER.toggle_pause,
        'XF86AudioMedia': exaile.gui.main.window.present,
        'XF86AudioRewind': lambda: PLAYER.seek(0),
    }[key]()



# Copyright (C) 2017 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from gi.repository import GLib
from gi.repository import Gtk

from xl import event, providers
from xl.nls import gettext as _

from xlgui.guiutil import GtkTemplate
from xlgui.widgets import menu

import threading


class DeveloperPlugin:
    """
    Shows useful information for Exaile developers
    """

    def enable(self, exaile):
        self.lock = threading.RLock()

        self.exaile = exaile
        self.window = None
        self.menu = None

        # Event data storage
        self.events = {'__all': 0}
        self.capture_events = True

        # subscribe for all events
        event.add_callback(self.on_events)

    def disable(self, exaile):
        self.teardown(exaile)

    def teardown(self, exaile):
        event.remove_callback(self.on_events)
        if self.window:
            self.window.destroy()
        if self.menu:
            providers.unregister('menubar-tools-menu', self.menu)

    def on_gui_loaded(self):

        # add a thing to the view menu
        self.menu = menu.simple_menu_item(
            'developer', '', _('Developer Tools'), callback=self.on_view_menu
        )

        providers.register('menubar-tools-menu', self.menu)

    def on_view_menu(self, widget, name, parent, context):
        if self.window:
            self.window.present()
        else:
            self.window = DeveloperWindow(self.exaile.gui.main.window, self)

            def _delete(w, e):
                self.window = None

            self.window.connect('delete-event', _delete)
            self.window.show_all()

    #
    # Capture callback stuff
    #

    def clear_events(self):
        with self.lock:
            self.events.clear()
            self.events['__all'] = 0

    def pause_events(self, pause):
        with self.lock:
            self.capture_events = pause

    def on_events(self, etype, *args):
        with self.lock:
            if not self.capture_events:
                return
            self.events['__all'] += 1
            try:
                self.events[etype] += 1
            except KeyError:
                self.events[etype] = 1

    def get_event_data(self, last_count):
        with self.lock:
            all_count = self.events['__all']
            if all_count != last_count:
                return self.events.copy(), all_count


plugin_class = DeveloperPlugin


@GtkTemplate('developer_window.ui', relto=__file__)
class DeveloperWindow(Gtk.Window):

    __gtype_name__ = 'DeveloperWindow'

    (
        event_filter_entry,
        event_model_filter,
        event_tree,
        event_store,
    ) = GtkTemplate.Child.widgets(4)

    def __init__(self, parent, plugin):
        Gtk.Window.__init__(self)
        self.init_template()

        if parent:
            self.set_transient_for(parent)

        self.events_count = None
        self.plugin = plugin
        self.event_filter_text = ''

        # key: name, value: iter
        self.event_model_idx = {}

        self.event_model_filter.set_visible_func(self.on_event_filter_row)

        self.event_tree.set_search_entry(self.event_filter_entry)
        self.event_tree.connect(
            'start-interactive-search', lambda *a: self.event_filter_entry.grab_focus()
        )

        self.event_timeout_id = GLib.timeout_add(250, self.on_event_update)

    @GtkTemplate.Callback
    def on_clear_events(self, widget):
        self.plugin.clear_events()
        self.events_count = None
        self.event_model_idx.clear()
        self.event_store.clear()

    @GtkTemplate.Callback
    def on_delete(self, widget, event):
        GLib.source_remove(self.event_timeout_id)

    def on_event_filter_row(self, model, titer, unused):
        row = model.get(titer, 0)
        if row:
            return self.event_filter_text in row[0]

    @GtkTemplate.Callback
    def on_event_pause_toggle(self, widget):
        self.plugin.pause_events(widget.get_active())

    @GtkTemplate.Callback
    def on_event_filter_entry_changed(self, widget):
        self.event_filter_text = widget.get_text()
        self.event_model_filter.refilter()

    def on_event_update(self):
        '''Periodically updates the displayed list of events'''
        data = self.plugin.get_event_data(self.events_count)
        if data:
            events, self.events_count = data
            for name, count in events.items():
                titer = self.event_model_idx.get(name)
                if titer:
                    self.event_store[titer][1] = count
                else:
                    titer = self.event_store.append([name, count])
                    self.event_model_idx[name] = titer

        return True



# moodbar - Replace Exaile's seekbar with a moodbar
# Copyright (C) 2015, 2019  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


# Ideas taken from:
# - The original Moodbar visualization method in Amarok 1;
# - The previous Moodbar plugin included in Exaile <4, by Solyanov Michael.


import os.path

from gi.repository import Gdk, GLib

import xl.event
from xl.nls import gettext as _
import xl.player
import xl.xdg
import xlgui.guiutil

from .cache import MoodbarCache
from .generator import MoodbarGenerator
from . import prefs
from .widget import Moodbar


class MoodbarPlugin:
    def __init__(self):
        self.main_controller = self.preview_controller = None

    def enable(self, exaile):
        self.generator = MoodbarGenerator()
        self.generator.check()

        self.exaile = exaile
        self.cache = MoodbarCache(os.path.join(xl.xdg.get_cache_dir(), 'moods'))

        xl.event.add_ui_callback(
            self._on_preview_device_enabled, 'preview_device_enabled'
        )
        xl.event.add_ui_callback(
            self._on_preview_device_disabling, 'preview_device_disabling'
        )

    def on_gui_loaded(self):
        self.main_controller = MoodbarController(
            self, xl.player.PLAYER, self.exaile.gui.main.progress_bar
        )
        previewdevice = self.exaile.plugins.enabled_plugins.get('previewdevice', None)
        if previewdevice:
            self._on_preview_device_enabled('', previewdevice)

    def get_preferences_pane(self):
        return prefs

    def disable(self, exaile):
        if not self.main_controller:  # Disabled more than once or before gui_loaded
            return
        xl.event.remove_callback(
            self._on_preview_device_enabled, 'preview_device_enabled'
        )
        xl.event.remove_callback(
            self._on_preview_device_disabling, 'preview_device_disabling'
        )
        self.main_controller.destroy()
        if self.preview_controller:
            self.preview_controller.destroy()
        self.main_controller = self.preview_controller = None
        del self.exaile, self.cache, self.generator

    # Preview Device events

    def _on_preview_device_enabled(self, event, plugin, _=None):
        self.preview_controller = MoodbarController(
            self, plugin.player, plugin.progress_bar
        )

    def _on_preview_device_disabling(self, event, plugin, _=None):
        self.preview_controller.destroy()
        self.preview_controller = None


plugin_class = MoodbarPlugin


# TRANSLATORS: Time format for playback progress
def format_time(seconds, time_format=_("{minutes}:{seconds:02}")):
    seconds = int(round(seconds))
    minutes, seconds = divmod(seconds, 60)
    return time_format.format(minutes=int(minutes), seconds=seconds)


class MoodbarController:
    def __init__(self, plugin, player, orig_seekbar):
        self.plugin = plugin
        self.player = player
        self.orig_seekbar = orig_seekbar
        self.timer = self.seeking = False

        self.moodbar = moodbar = Moodbar()
        self._update_painter()
        self._update_tint()
        xlgui.guiutil.gtk_widget_replace(self.orig_seekbar, moodbar)
        moodbar.show()

        # TODO: If currently playing, this needs to run now as well:
        xl.event.add_ui_callback(
            self._on_playback_track_start, 'playback_track_start', self.player
        )
        xl.event.add_ui_callback(
            self._on_playback_track_end, 'playback_track_end', self.player
        )

        moodbar.add_events(
            Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON1_MOTION_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
        )
        moodbar.connect('button-press-event', self._on_moodbar_button_press)
        moodbar.connect('motion-notify-event', self._on_moodbar_motion)
        moodbar.connect('button-release-event', self._on_moodbar_button_release)

        xl.event.add_callback(self._on_option_set, 'plugin_moodbar_option_set')

    def destroy(self):
        xl.event.remove_callback(
            self._on_playback_track_end, 'playback_track_start', self.player
        )
        xl.event.remove_callback(
            self._on_playback_track_end, 'playback_track_end', self.player
        )
        if self.timer:
            GLib.source_remove(self.timer)
        assert self.orig_seekbar
        xlgui.guiutil.gtk_widget_replace(self.moodbar, self.orig_seekbar)
        self.moodbar.destroy()
        self.moodbar = None

    # Helpers to set moodbar state from prefs

    def _update_painter(self):
        use_waveform = xl.settings.get_option(
            prefs.UseWaveformPreference.name, prefs.UseWaveformPreference.default
        )
        self.moodbar.set_use_waveform(use_waveform)

    def _update_tint(self):
        use_tint = xl.settings.get_option(
            prefs.UseTintPreference.name, prefs.UseTintPreference.default
        )
        if use_tint:
            tint_s = xl.settings.get_option(
                prefs.TintPreference.name, prefs.TintPreference.default
            )
            tint = Gdk.RGBA()
            tint.parse(tint_s)
        else:
            tint = None
        self.moodbar.set_tint(tint)

    # Playback events

    def _on_playback_track_start(self, event, player, track):
        cache = self.plugin.cache
        uri = player.current.get_loc_for_io()
        data = cache.get(uri) if cache else None
        self.moodbar.set_mood(data)
        self._on_timer()
        self.timer = GLib.timeout_add_seconds(1, self._on_timer)
        if not data and uri.startswith('file://'):

            def callback(uri, data):
                if cache:
                    cache.put(uri, data)
                self.moodbar.set_mood(data)

            self.plugin.generator.generate_async(uri, callback)

    def _on_timer(self):
        assert self.moodbar
        try:
            total_time = self.player.current.get_tag_raw('__length')
        except AttributeError:  # No current track
            return False
        current_time = self.player.get_time()
        if total_time:
            time_format = dict(
                current=format_time(current_time),
                remaining=format_time(total_time - current_time),
            )
            # TRANSLATORS: Format for playback progress text
            self.moodbar.set_text(_("{current} / {remaining}").format(**time_format))
            if not self.seeking:
                self.moodbar.seek_position = current_time / total_time
        else:
            self.moodbar.set_text(format_time(current_time))
        return True

    def _on_playback_track_end(self, event, player, track):
        if self.timer:
            GLib.source_remove(self.timer)
        self.timer = None
        self.moodbar.set_mood(None)
        self.moodbar.seek_position = None
        self.moodbar.set_text(None)

    # Mouse events

    def _on_moodbar_button_press(self, moodbar, event):
        """
        :type moodbar: Moodbar
        :type event: Gdk.Event
        """
        if event.button == Gdk.BUTTON_PRIMARY:
            self.seeking = True
            self._on_moodbar_motion(moodbar, event)

    def _on_moodbar_motion(self, moodbar, event):
        """
        :type moodbar: Moodbar
        :type event: Gdk.Event
        """
        if self.seeking:
            w = moodbar.get_allocation().width
            x = event.get_coords()[0]
            x = max(0, min(x, w))
            moodbar.seek_position = x / w

    def _on_moodbar_button_release(self, moodbar, event):
        """
        :type moodbar: Moodbar
        :type event: Gdk.Event
        """
        if event.button == Gdk.BUTTON_PRIMARY and self.seeking:
            self.player.set_progress(moodbar.seek_position)
            self.seeking = False

    # Preferences events

    def _on_option_set(self, _event, _settings, option):
        if option == prefs.UseWaveformPreference.name:
            self._update_painter()
        elif option in (prefs.UseTintPreference.name, prefs.TintPreference.name):
            self._update_tint()
        else:
            assert False



# Copyright (C) 2019 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging
import os
from typing import Dict

logger = logging.getLogger(__name__)

__version__ = "devel"

# We need the local hack for OSX bundled apps, so we depend on the main script
# to set the environment variable correctly instead of trying to infer an
# absolute path
# exaile_dir = os.path.split(os.path.dirname(os.path.realpath(__file__)))[0]
exaile_dir = os.environ['EXAILE_DIR']


def get_current_version():
    """
    Get the latest version identifier for the branch contained in the
    current directory. Returns None if the directory is not a branch or
    the version identifier cannot be found.
    """
    import subprocess

    try:
        return (
            subprocess.check_output(
                ['git', 'describe', '--tags', '--abbrev=0'], stderr=subprocess.DEVNULL
            )
            .strip()
            .decode('utf-8')
        )
    except (subprocess.CalledProcessError, OSError):
        return None


def get_current_revision():
    """
    Get the latest revision identifier for the branch contained in the
    current directory. Returns None if the directory is not a branch or
    the revision identifier cannot be found.
    """
    import subprocess

    try:
        return (
            subprocess.check_output(
                ['git', 'rev-parse', '--short=7', 'HEAD'], stderr=subprocess.DEVNULL
            )
            .strip()
            .decode('utf-8')
        )
    except (subprocess.CalledProcessError, OSError):
        return None


if "DIST_VERSION" in os.environ:
    __version__ = os.environ['DIST_VERSION']
elif os.path.exists(os.path.join(exaile_dir, ".git")):
    version = get_current_version()
    if version is not None:
        __version__ = version
    revision = get_current_revision()
    if revision is not None:
        __version__ += "+" + revision

__external_versions__: Dict[str, str] = {}


def register(name, version):
    '''Registers versions of external components for diagnostic purposes'''
    if name not in __external_versions__:
        __external_versions__[name] = version
        logger.info("Using %s %s", name, version)
